from urllib.parse import urlparse
from dataclasses import dataclass, field
from collections import defaultdict
from tarfile import TarFile
from minio import Minio
from minio.helpers import check_bucket_name
from minio.error import S3Error
from typing import List
from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache

import subprocess
import struct
import traceback
import ipaddress
import tempfile
import requests
import json
import re
import os
import time

from . import config
from .logger import logger


@dataclass
class MinIOPolicyStatement:
    # NOTE: Using the default factory to avoid mutable defaults
    buckets: List[str] = field(default_factory=lambda: ["*"])
    actions: List[str] = field(default_factory=lambda: ["s3:*"])
    allow: bool = True

    # Conditions
    source_ips: List[str] = None
    not_source_ips: List[str] = None


def generate_policy(statements):
    def nesteddict():
        return defaultdict(nesteddict)

    rendered_statements = []
    for s in statements:
        resources = [f"arn:aws:s3:::{b}" for b in s.buckets]
        resources.extend([f"arn:aws:s3:::{b}/*" for b in s.buckets])

        statement = {
            'Action': s.actions,
            'Effect': "Allow" if s.allow else "Deny",
            'Resource': resources,
        }

        conditions = nesteddict()
        if s.source_ips and len(s.source_ips) > 0:
            conditions["IpAddress"]["aws:SourceIp"] = s.source_ips
        if s.not_source_ips and len(s.not_source_ips) > 0:
            conditions["NotIpAddress"]["aws:SourceIp"] = s.not_source_ips
        if len(conditions) > 0:
            statement["Condition"] = conditions

        rendered_statements.append(statement)

    return {
        "Version": "2012-10-17",
        "Statement": rendered_statements
    }


# TODO: rename the methods to be on the form $object_$operation
# to make auto-completion work more efficiently.
class MinioClient():
    @classmethod
    def __log_mcli_error(cls, e):
        logger.error(f"""Call failed:
Command output:
{e.output}

Backtrace:
{traceback.format_exc()}
""")

    def __mcli_exec(self, args):
        assert self.alias is not None

        try:
            return subprocess.check_output(["mcli", "--no-color"] + args)
        except subprocess.CalledProcessError as e:  # pragma: nocover
            self.__log_mcli_error(e)
            raise e from None

    def __init__(self,
                 url=config.MINIO_URL,
                 user=config.MINIO_ROOT_USER,
                 secret_key=config.MINIO_ROOT_PASSWORD,
                 alias=config.MINIO_ADMIN_ALIAS,
                 artifact_cache_root=config.EXECUTOR_ARTIFACT_CACHE_ROOT):
        self.url = url
        self.user = user
        self.secret_key = secret_key
        self.alias = alias
        self.artifact_cache_root = artifact_cache_root

        self._client = Minio(
            endpoint=urlparse(url).netloc,
            access_key=user,
            secret_key=secret_key,
            secure=False,
        )

        # Some operations can only be used using the commandline tool,
        # so initialize it
        if alias is not None:
            try:
                subprocess.check_call(
                    ["mcli", "--no-color", "alias", "set", self.alias, url,
                     self.user, self.secret_key])
            except subprocess.CalledProcessError as e:  # pragma: nocover
                self.__log_mcli_error(e)
                raise ValueError("Invalid MinIO credentials") from None

    def remove_alias(self):
        if self.alias is not None:
            self.__mcli_exec(["alias", "rm", self.alias])

    def is_local_url(self, url):
        return url.startswith(f"{self.url}/")

    def save_boot_artifact(self, remote_artifact_url, minio_object_name, log_callback=logger.info):
        minio_bucket_name = 'boot'

        disable_cache = self.artifact_cache_root is None
        if not disable_cache:
            try:
                os.makedirs(self.artifact_cache_root, exist_ok=True)
            except PermissionError:  # pragma: no cover
                log_callback(f"Unable to create \"{self.artifact_cache_root}\", "
                             "permission denied.  Disabling caching of artifacts...")
                disable_cache = True

        # For timing the request.get
        start_time = time.time()
        if disable_cache:
            resp = requests.get(remote_artifact_url)

        else:
            session = CacheControl(requests.Session(), cache=FileCache(self.artifact_cache_root))
            resp = session.get(remote_artifact_url)

        # print some additional statistics when the fetch was with caching
        # disabled or was a cache miss
        from_cache = disable_cache is False and resp.from_cache
        size_mb = len(resp.content) / 1024**2
        if not from_cache:
            # note: don't round total_time here else it might cause a #DE when
            # calculating avg speed
            total_time = time.time() - start_time
            avg_speed = size_mb / total_time
            log_callback(f"Downloaded {'and cached ' if not disable_cache else ''}"
                         f"the boot artifact in {round(total_time, 1)}s "
                         f"({round(size_mb, 2)} MB, at {round(avg_speed, 2)} MB/s)")
        else:
            log_callback(f"Fetched the boot artifact from cache ({round(size_mb, 2)} MB)")

        resp.raise_for_status()

        with tempfile.NamedTemporaryFile("wb") as temp_download_area:
            temp_download_area.write(resp.content)
            temp_download_area.flush()
            self._client.fput_object(minio_bucket_name, minio_object_name,
                                     temp_download_area.name)

    def _build_mc_attrs_str(self, tarball_member):
        ti = tarball_member.get_info()
        m = tarball_member
        # Unfortunately mc uses non-portable Go encodings in their public protocol:
        #   https://pkg.go.dev/io/fs#FileMode
        # They shouldn't do that, instead only the least
        # significant 9 bits of the go FileMode structure should be
        # exposed (the UNIX permissions). This is reality however, so
        # we need to deal with the fact they don't do it like that,
        # and try our best.  Currently mc ignores the non-standard
        # bits at least, but be a little defensive in case they change
        # their minds.
        gomode = 0
        if m.isdir():
            gomode |= 1 << 31
        if m.islnk():
            gomode |= 1 << 27
        if m.isdev():
            gomode |= 1 << 26
        if m.isfifo():
            gomode |= 1 << 25
        if m.ischr():
            gomode |= 1 << 21
        if m.isreg():
            gomode |= 1 << 15
        # Low 9-bits for file mode
        gomode |= m.mode
        # Yes, they pack is little-endian over the network.
        mode = int.from_bytes(struct.pack('<I', gomode), byteorder='little')
        return f"gid:{ti['gid']}/gname:{ti['gname']}/mode:{mode}/mtime:{int(ti['mtime'])}/uid:{ti['uid']}/uname:{ti['uname']}"  # noqa

    def extract_archive(self, archive_fileobj, bucket_name):
        with TarFile.open(fileobj=archive_fileobj, mode='r') as archive:
            while (member := archive.next()) is not None:
                # Ignore everything that isn't a file
                if not member.isfile():
                    continue
                metadata = {
                    'X-Amz-Meta-Mc-Attrs': self._build_mc_attrs_str(member)
                }
                self._client.put_object(bucket_name, member.name, archive.extractfile(member),
                                        member.size, num_parallel_uploads=1, metadata=metadata)

    def make_bucket(self, bucket_name):
        try:
            self._client.make_bucket(bucket_name)
        except S3Error:
            raise ValueError("The bucket already exists") from None

    def bucket_exists(self, bucket_name):
        return self._client.bucket_exists(bucket_name)

    # NOTE: Using minioclient's remove_bucket requires first to empty the
    # bucket. Use the CLI version for now.
    def remove_bucket(self, bucket_name):
        self.__mcli_exec(["rb", "--force", f'{self.alias}/{bucket_name}'])

    def add_user(self, user_id, password):
        self.__mcli_exec(["admin", "user", "add", self.alias, user_id, password])

    def remove_user(self, user_id):
        self.__mcli_exec(["admin", "user", "remove", self.alias, user_id])

    def groups_user_is_in(self, user_id=None):
        if user_id is None:
            user_id = self.user

        try:
            output = self.__mcli_exec(["--json", "admin", "user", "info", self.alias, user_id])
        except subprocess.CalledProcessError:  # pragma: nocover
            raise ValueError("Failed to query information about the user") from None

        # Parse the output
        groups = set()
        for g in json.loads(output).get('memberOf', []):
            if isinstance(g, str):
                # Original format: List of strings
                groups.add(g)
            elif isinstance(g, dict):
                # Current format: List of dicts
                if name := g.get("name"):
                    groups.add(name)

        return groups

    def add_user_to_group(self, user_id, group_name):
        self.__mcli_exec(["admin", "group", "add", self.alias, group_name, user_id])

    def apply_user_policy(self, policy_name, user_id, policy_statements):
        with tempfile.NamedTemporaryFile(suffix='json') as f:
            policy = generate_policy(policy_statements)
            f.write(json.dumps(policy).encode())
            f.flush()

            self.__mcli_exec(["admin", "policy", "create", self.alias, policy_name, f.name])

            try:
                self.__mcli_exec(["--json", "admin", "policy", "attach", self.alias, policy_name, "--user", user_id])
            except subprocess.CalledProcessError as e:
                data = json.loads(e.output)
                error_code = data.get("error", {}).get("cause", {}).get("error", {}).get("Code")
                if error_code not in ["XMinioPolicyAlreadyAttached"]:
                    raise ValueError(f"Applying policy failed: Error: {error_code}") from None

    def remove_user_policy(self, policy_name, user_id):
        self.__mcli_exec(["admin", "policy", "detach", self.alias, policy_name, "--user", user_id])
        self.__mcli_exec(["admin", "policy", "remove", self.alias, policy_name])

    @classmethod
    def create_valid_bucket_name(cls, base_name):
        # Bucket names can consist only of lowercase letters, numbers, dots (.), and hyphens (-)
        name = base_name.lower()
        name = re.sub(r'[^a-z0-9\-\.]', '-', name)

        # Bucket name must not contain invalid successive chars ['..', '.-', '-.']
        # NOTE: The transformation is repeated until no such sequence may be found
        while True:
            name_before_sub = name
            name = re.sub(r'(\.\.|\.-|-\.)', '-', name)
            if name == name_before_sub:
                break

        # Bucket names must be between 3 and 63 characters long.
        if len(name) < 3:
            name = "b--" + name

        # Bucket names must begin and end with a letter or number.
        # Bucket names can't begin with xn-- (for buckets created after February 2020)
        if name.startswith('xn--') or name[0] == '.' or name[0] == '-':
            name = 'x' + name

        # Bucket names must not be formatted as an IP address (for example, 192.168.5.4)
        try:
            ipaddress.ip_address(name)

            # The name is an ip address, add a prefix!
            name = "ip-" + name
        except ValueError:
            # The name isn't an IP address, all is good!
            pass

        # Bucket names must be between 3 and 63 characters long.
        name = name[0:63]

        # Bucket names must begin and end with a letter or number.
        if name[-1] == '.' or name[-1] == '-':
            name = name[0:62] + 'x'

        # Do the final checks
        check_bucket_name(name)

        return name



from dataclasses import field, fields
from typing import Dict, List, Optional

from jinja2 import Template, ChainableUndefined
from pydantic.dataclasses import dataclass
import usb.core
import yaml

from .job import DeploymentState
from . import config


@dataclass
class USBMatcher:
    idVendor: Optional[int | List[int]] = None
    idProduct: Optional[int | List[int]] = None
    iManufacturer: Optional[str | List[str]] = None
    iProduct: Optional[str | List[str]] = None
    iSerialNumber: Optional[str | List[str]] = None

    def matches(self, device: 'usb.core.Device'):
        for f in fields(self):
            if expected_values := getattr(self, f.name, None):
                if not isinstance(expected_values, list):
                    expected_values = [expected_values]

                if getattr(device, f.name) not in expected_values:
                    return False

        return True


@dataclass
class FastbootDeviceMatcher:
    # NOTE: All the specified values must be a match to be considered a valid match
    usb: Optional[USBMatcher] = None
    variables: Optional[Dict[str, str | List[str]]] = field(default_factory=dict)  # The values are regular expressions

    def matches(self, fbdev: 'FastbootDevice'):
        if self.usb and not self.usb.matches(fbdev.device):
            return False

        # Match the expected variables
        fbdev_vars = fbdev.variables
        for name, expected_values in self.variables.items():
            if not isinstance(expected_values, list):
                expected_values = [expected_values]

            if fbdev_vars.get(name) not in expected_values:
                return False

        return True


@dataclass
class FastbootDevice:
    match: FastbootDeviceMatcher
    defaults: DeploymentState


@dataclass
class BootsDB:
    fastboot: Optional[Dict[str, FastbootDevice]] = field(default_factory=dict)

    @classmethod
    def render_with_resources(cls, db_str, machine=None, **kwargs):
        template_params = {
            **{k.lower(): v for k, v in config.job_environment_vars().items()},
            "machine": machine.safe_attributes if machine else None,
            **kwargs,
        }

        rendered_db_str = Template(db_str, undefined=ChainableUndefined).render(**template_params)
        return cls(**yaml.safe_load(rendered_db_str))

    @classmethod
    def from_path(cls, filepath, **kwargs):
        with open(filepath, "r") as f_template:
            db_str = f_template.read()
            return cls.render_with_resources(db_str, **kwargs)



import argparse
import sys


# Entrypoint for the executor script
def run():
    parser = argparse.ArgumentParser(prog='Executor Server')
    subparsers = parser.add_subparsers(help='Desired action to perform. Default: server', dest='action')

    # Server
    subparsers.add_parser("server", help='Start running the executor as a server')

    # Run-job
    parser_job = subparsers.add_parser("run-job", help='Start running a job')
    parser_job.add_argument('-c', '--config', type=argparse.FileType('r'), default=sys.stdin,
                            help="The job description to execute")
    parser_job.add_argument('-s', '--socket', required=True,
                            help="Path to the unix socket to use to communicate")
    parser_job.add_argument('-l', '--lock-file', required=True,
                            help="Path to the lock file to use")

    # Parse the command line
    args = parser.parse_args()

    # Execute the right entrypoint
    if args.action is None or args.action == "server":
        from .app import run
        run()
    elif args.action == "run-job":
        from .executor import run
        run(config_f=args.config, socket_path=args.socket, lock_path=args.lock_file)



from unittest.mock import patch, MagicMock
from server.boots import BootService, BootConfig, IPXE_BINARIES, BootsDHCPD
from server.dhcpd import CPUArch, Firmware, BootProtocol, MacAddress
import server.config as config
import pytest


@patch("server.boots.BootsTFTPD")
@patch("server.boots.BootsDHCPD")
@patch("server.boots.provision_ipxe_dut_clients")
def create_boot_service(tmp_path, mock_provisioner, mock_dhcpd, mock_tftpd):
    paths = {
        'TFTP_DIR': f'{tmp_path}/tftp',
    }
    service = BootService(MagicMock(),
                          private_interface='br0',
                          config_paths=paths)

    if not config.BOOTS_DISABLE_SERVERS:
        mock_dhcpd.assert_called_with(service, "DHCP Server", 'br0')
        mock_tftpd.assert_called_with(service, "TFTP Server", paths["TFTP_DIR"], 'br0')

    return service


def test_boot_service(tmp_path):
    create_boot_service(tmp_path)


def test_boot_service_disabled(tmp_path):
    config.BOOTS_DISABLE_SERVERS = True
    try:
        create_boot_service(tmp_path)
    finally:
        config.BOOTS_DISABLE_SERVERS = False


@patch("server.boots.DHCPD.__init__")
def test_dhcpd_static_clients(dhcpd_mock):
    boots = MagicMock()
    dhcp = BootsDHCPD(boots, name="toto", interface="br0")

    m_mac = "00-01-02-03-04-05"
    m_ipaddr = "10.42.0.42"
    m_hostname = "machine1"

    boots.mars.known_machines = [MagicMock(mac_address=m_mac, ip_address=m_ipaddr, full_name=m_hostname)]
    assert dhcp.static_clients == [{'mac_addr': MacAddress(m_mac), 'ipaddr': m_ipaddr, 'hostname': m_hostname}]


@patch("server.boots.DHCPD.__init__")
def test_dhcpd_boot_target(dhcpd_mock):
    dhcp = BootsDHCPD(MagicMock(), name="toto", interface="br0")
    dhcp.logger = MagicMock()
    dhcp.get_or_assign_ip_for_client = MagicMock(return_value=("10.0.0.42", "new client"))

    dhcpd_mock.assert_called_once_with(dhcp, interface="br0")

    # Legacy X86 boot
    request = MagicMock(architecture=CPUArch.X86, firmware=Firmware.BIOS,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == IPXE_BINARIES['IPXE_I386_MBR_FILENAME']
    dhcp.boots.mars.machine_discovered.assert_called_once_with(
        {
            "id": "00:01:02:03:04:05",
            "mac_address": "00:01:02:03:04:05",
            "base_name": "x86-bios-tftp",
            "ip_address": "10.0.0.42",
            "tags": []
        }, update_if_already_exists=False)

    # X86 UEFI boot
    request = MagicMock(architecture=CPUArch.X86, firmware=Firmware.UEFI,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == IPXE_BINARIES['IPXE_I386_EFI_FILENAME']

    # X86_64 UEFI boot
    request = MagicMock(architecture=CPUArch.X86_64, firmware=Firmware.UEFI,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == IPXE_BINARIES['IPXE_X86_64_EFI_FILENAME']

    # ARM32 UEFI boot
    request = MagicMock(architecture=CPUArch.ARM32, firmware=Firmware.UEFI,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == IPXE_BINARIES['IPXE_ARM32_EFI_FILENAME']

    # ARM64 UEFI boot
    request = MagicMock(architecture=CPUArch.ARM64, firmware=Firmware.UEFI,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == IPXE_BINARIES['IPXE_ARM64_EFI_FILENAME']

    # ARM32 U-Boot boot
    request = MagicMock(architecture=CPUArch.ARM32, firmware=Firmware.UBOOT,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == "/dyn/bootloader/u-boot/ARM32/00:01:02:03:04:05/boot.scr"

    # ARM64 U-Boot boot
    request = MagicMock(architecture=CPUArch.ARM64, firmware=Firmware.UBOOT,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) == "/dyn/bootloader/u-boot/ARM64/00:01:02:03:04:05/boot.scr"

    # Unsupported Architecture
    request = MagicMock(architecture=CPUArch.RISCV32, firmware=Firmware.BIOS,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) is None
    dhcp.logger.error.assert_called_once_with("Unsupported architecture 'RISCV32'")
    dhcp.logger.error.reset_mock()

    # Unsupported Architecture for BIOS
    request = MagicMock(architecture=CPUArch.ARM32, firmware=Firmware.BIOS,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) is None
    dhcp.logger.error.assert_called_once_with("Unsupported BIOS architecture 'ARM32'")
    dhcp.logger.error.reset_mock()

    # Unsupported Protocol
    request = MagicMock(architecture=CPUArch.X86, firmware=Firmware.UEFI,
                        protocol=BootProtocol.HTTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) is None
    dhcp.logger.error.assert_called_once_with("Unsupported protocol 'HTTP'")
    dhcp.logger.error.reset_mock()

    # Unsupported Firmware
    request = MagicMock(architecture=CPUArch.X86, firmware=Firmware.RPI,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"))
    assert dhcp.boot_target(request) is None
    dhcp.logger.error.assert_called_once_with("Unsupported firmware type 'RPI'")
    dhcp.logger.error.reset_mock()

    # iPXE request on X86/BIOS/TFTP
    request = MagicMock(architecture=CPUArch.X86, firmware=Firmware.BIOS,
                        protocol=BootProtocol.TFTP, mac_addr=MacAddress("00-01-02-03-04-05"),
                        user_class="iPXE")
    assert (dhcp.boot_target(request) ==
            "http://ci-gateway/boot/00:01:02:03:04:05/boot.ipxe?platform=pcbios&buildarch=i386")


def test_ipxe_boot_script(tmp_path, monkeypatch):
    service = create_boot_service(tmp_path)

    # Set the default boot configuration
    expected_boot_cfg = BootConfig(kernel="default_kernel",
                                   initrd="default_initrd",
                                   cmdline="default_cmdline")
    monkeypatch.setenv("BOOTS_DEFAULT_KERNEL", expected_boot_cfg.kernel)
    monkeypatch.setenv("BOOTS_DEFAULT_INITRD", expected_boot_cfg.initrd)
    monkeypatch.setenv("BOOTS_DEFAULT_CMDLINE", expected_boot_cfg.cmdline)

    # Check that when no machines are specified, we use the default boot configuration
    service._gen_ipxe_boot_script = MagicMock()
    service.ipxe_boot_script(machine=None)
    service._gen_ipxe_boot_script.assert_called_once_with(expected_boot_cfg, platform=None)

    # Check that we use the bootconfig when defined
    service._gen_ipxe_boot_script = MagicMock()
    machine = MagicMock()
    platform = MagicMock()
    buildarch = MagicMock()
    service.ipxe_boot_script(machine=machine, platform=platform, buildarch=buildarch)
    machine.boot_config_query.assert_called_with(platform=platform, buildarch=buildarch, bootloader="ipxe")


class UbootTestingMars(MagicMock):
    def get_machine_by_id(self, machine_id, raise_if_missing=False):
        if machine_id == "none":
            return None

        return self.machine_mock


def test_uboot_boot_script(tmp_path, monkeypatch):
    service = create_boot_service(tmp_path)

    service.mars = UbootTestingMars()

    # Set the default boot configuration
    expected_boot_cfg = BootConfig(kernel="default_kernel",
                                   initrd="default_initrd",
                                   cmdline="default_cmdline")
    monkeypatch.setenv("BOOTS_DEFAULT_KERNEL", expected_boot_cfg.kernel)
    monkeypatch.setenv("BOOTS_DEFAULT_INITRD", expected_boot_cfg.initrd)
    monkeypatch.setenv("BOOTS_DEFAULT_CMDLINE", expected_boot_cfg.cmdline)

    # Check that when no machines is invalid, we use the default boot configuration
    service._gen_uboot_boot_script = MagicMock(return_value="default_script")
    service.uboot_boot_script(machine_id="none")
    service._gen_uboot_boot_script.assert_called_once_with(expected_boot_cfg, platform="uboot")

    # Check that we use the bootconfig when defined
    service._gen_uboot_boot_script = MagicMock(return_value="default_script")
    machine = MagicMock()
    service.mars.machine_mock = machine
    buildarch = MagicMock()
    service.uboot_boot_script(machine_id="00-01-02-03-04-05", buildarch=buildarch)
    machine.boot_config_query.assert_called_with(platform="uboot", buildarch=buildarch, bootloader="uboot")

    for name in [CPUArch.ARM32.name, CPUArch.ARM64.name,
                 CPUArch.X86.name, CPUArch.X86_64.name,
                 CPUArch.RISCV32.name, CPUArch.RISCV64.name]:
        service.uboot_boot_script(machine_id="00-01-02-03-04-05", buildarch=name)
        machine.boot_config_query.assert_called_with(platform="uboot", buildarch=name, bootloader="uboot")


def test_platform_cmdline():
    assert BootService._platform_cmdline() == ""
    assert BootService._platform_cmdline(platform="efi") == "initrd=initrd"
    assert BootService._platform_cmdline(platform="pcbios") == ""


def test_gen_ipxe_boot_script():
    BootService._platform_cmdline = MagicMock(return_value="platform_args")

    script = BootService._gen_ipxe_boot_script(BootConfig(kernel="/kernel", initrd="/initrd", cmdline="cmdline"))
    assert "kernel /kernel platform_args cmdline\n" in script
    assert "initrd --name initrd /initrd\n" in script
    assert "boot\n" in script


def test_gen_uboot_boot_script():
    BootService._platform_cmdline = MagicMock(return_value="platform_args")

    script = BootService._gen_uboot_boot_script(BootConfig(kernel="/kernel", initrd="/initrd", cmdline="cmdline",
                                                           dtb="/dtb"))
    assert "tftp ${kernel_addr_r} minio/kernel\n" in script
    assert "tftp ${ramdisk_addr_r} minio/initrd\n" in script
    assert "tftp ${fdt_addr} minio/dtb\n" in script
    assert "booti ${kernel_addr_r} ${ramdisk_addr_r}:${ramdisk_size} ${fdt_addr}\n" in script


def test_default_boot_config(monkeypatch):
    # Check what happens without boot configs
    with pytest.raises(ValueError, match="No default kernel found.*"):
        BootConfig.defaults()
    monkeypatch.setenv("BOOTS_DEFAULT_KERNEL", "default_kernel")
    with pytest.raises(ValueError, match="No default initramfs found.*"):
        BootConfig.defaults()
    monkeypatch.setenv("BOOTS_DEFAULT_INITRD", "default_initrd")
    with pytest.raises(ValueError, match="No default kernel command line found.*"):
        BootConfig.defaults()
    monkeypatch.setenv("BOOTS_DEFAULT_CMDLINE", "default_cmdline")

    # Basic test
    b = BootConfig.defaults()
    assert b == BootConfig(kernel="default_kernel", initrd="default_initrd", cmdline="default_cmdline")

    # Test more complex configurations
    monkeypatch.setenv("BOOTS_DEFAULT_IPXE_X86_64_EFI_KERNEL", "ipxe_x86_64_efi_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_X86_64_EFI_KERNEL", "x86_64_efi_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_IPXE_X86_64_KERNEL", "ipxe_x86_64_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_X86_64_KERNEL", "x86_64_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_IPXE_KERNEL", "ipxe_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_X86_64_INITRD", "ipxe_x86_64_initrd")
    monkeypatch.setenv("BOOTS_DEFAULT_IPXE_ARM64_PCBIOS_KERNEL", "ipxe_arm64_pcbios_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_ARM64_EFI_INITRD", "arm64_efi_initrd")
    monkeypatch.setenv("BOOTS_DEFAULT_ARM64_CMDLINE", "arm64_cmdline")

    b = BootConfig.defaults(bootloader="ipxe", buildarch="x86_64", platform="efi")
    assert b == BootConfig(kernel="ipxe_x86_64_efi_kernel", initrd="ipxe_x86_64_initrd", cmdline="default_cmdline")

    b = BootConfig.defaults(buildarch="x86_64", platform="efi")
    assert b == BootConfig(kernel="x86_64_efi_kernel", initrd="ipxe_x86_64_initrd", cmdline="default_cmdline")

    b = BootConfig.defaults(buildarch="x86_64")
    assert b == BootConfig(kernel="x86_64_kernel", initrd="ipxe_x86_64_initrd", cmdline="default_cmdline")

    b = BootConfig.defaults(bootloader="ipxe")
    assert b == BootConfig(kernel="ipxe_kernel", initrd="default_initrd", cmdline="default_cmdline")

    # Check that BOOTS_DEFAULT_IPXE_ARM64_PCBIOS_KERNEL is not accidentally picked
    b = BootConfig.defaults(bootloader="uboot", buildarch="arm64", platform="efi")
    assert b == BootConfig(kernel="default_kernel", initrd="arm64_efi_initrd", cmdline="arm64_cmdline")


def test_fixup_missing_fields_with_defaults(monkeypatch):
    def empty_boot_cfg():
        return BootConfig(kernel=None, initrd=None, cmdline=None)

    # Check what happens without boot configs
    with pytest.raises(ValueError, match="No default kernel found.*"):
        empty_boot_cfg().fixup_missing_fields_with_defaults()
    monkeypatch.setenv("BOOTS_DEFAULT_KERNEL", "default_kernel")
    with pytest.raises(ValueError, match="No default initramfs found.*"):
        empty_boot_cfg().fixup_missing_fields_with_defaults()
    monkeypatch.setenv("BOOTS_DEFAULT_INITRD", "default_initrd")

    # Basic test
    bcfg = empty_boot_cfg()
    bcfg.fixup_missing_fields_with_defaults()
    assert bcfg == BootConfig(kernel="default_kernel", initrd="default_initrd", cmdline=None)

    # Test more complex configurations
    monkeypatch.setenv("BOOTS_DEFAULT_IPXE_X86_64_EFI_KERNEL", "ipxe_x86_64_efi_kernel")
    monkeypatch.setenv("BOOTS_DEFAULT_X86_64_INITRD", "ipxe_x86_64_initrd")
    monkeypatch.setenv("BOOTS_DEFAULT_CMDLINE", "cmdline")

    # Make sure both the kernel and initrd get updated, but not the commandline
    bcfg = empty_boot_cfg()
    bcfg.fixup_missing_fields_with_defaults(bootloader="ipxe", buildarch="x86_64", platform="efi")
    assert bcfg == BootConfig(kernel="ipxe_x86_64_efi_kernel", initrd="ipxe_x86_64_initrd", cmdline=None)



#!/usr/bin/env python3

from datetime import datetime

import multiprocessing
import traceback
import signal
import socket
import flask
import json
import time
import sys

from waitress import serve

from .dut import DUT, JobRequest, DUTState, SergentHartman, SergentHartmanState, InvalidTarballFile
from .executor import JobConsole
from .mars import Mars
from .minioclient import MinioClient
from .boots import BootService
from .message import JobStatus
from .pdu import PDUPort, PDUState
from .pdu.daemon import PDUDaemon, AsyncPDU
from .socketactivation import get_sockets_by_name
from . import config


class CustomJSONEncoder(flask.json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, JobStatus):
            return obj.name
        elif isinstance(obj, SergentHartmanState):
            return obj.name
        elif isinstance(obj, SergentHartman):
            return {
                "state": obj.state,
                "is_active": obj.is_active,
                "is_registered": obj.is_machine_registered,
                "boot_loop_counts": obj.boot_loop_counts,
                "qualifying_rate": obj.qualifying_rate,
                "current_loop_count": obj.cur_loop,
                "statuses": dict([(s.name, val) for s, val in obj.statuses.items()]),
            }
        elif isinstance(obj, DUTState):
            return obj.name
        elif isinstance(obj, DUT):
            dut = {
                "id": obj.id,
                "state": obj.state,
                "ready_for_service": obj.ready_for_service,
                "has_pdu_assigned": obj.pdu_port is not None,
                "local_tty_device": obj.local_tty_device,
                "tags": list(obj.tags),
                "manual_tags": list(obj.manual_tags),
                "base_name": obj.base_name,
                "full_name": obj.full_name,
                "mac_address": obj.mac_address,
                "ip_address": obj.ip_address,
                "is_retired": obj.is_retired,
                "training": obj.sergent_hartman,
                "pdu": {
                    "name": obj.pdu,
                    "port_id": obj.pdu_port_id
                },
                "pdu_off_delay": obj.pdu_off_delay,
                "comment": obj.comment
            }

            if salad := JobConsole.salad_request(obj.id):
                dut["logs_endpoint"] = {
                    "host": "ci-gateway",
                    "port": salad["tcp_port_logs"],
                }
            return dut
        elif isinstance(obj, AsyncPDU):
            return {
                "ports": {p.port_id: p for p in obj.ports},
                "state": obj.state.name,
                "error": obj.error,

            }
        elif isinstance(obj, PDUPort):
            return {
                "label": obj.label,
                "min_off_time": obj.min_off_time,
                "state": obj.last_known_state.name,
                "last_shutdown": obj.last_shutdown,
                "last_polled": obj.last_polled,
                "reserved": obj.reserved
            }

        return super().default(obj)


class CustomJSONProvider(flask.json.provider.JSONProvider):
    def dumps(self, obj, *args, **kwargs):
        return json.dumps(obj, cls=CustomJSONEncoder, *args, **kwargs)

    def loads(self, s, **kwargs):
        return json.loads(s)


def find_pdu(pdu_name):
    with app.app_context():
        mars = flask.current_app.mars
    return mars.get_pdu_by_name(pdu_name, raise_if_missing=True)


def find_pdu_port(pdu_name, port_id):
    with app.app_context():
        mars = flask.current_app.mars
    return mars.get_pdu_port_by_name(pdu_name, port_id, raise_if_missing=True)


app = flask.Flask(__name__)
app.json = CustomJSONProvider(app)


@app.errorhandler(ValueError)
def handle_valueError_exception(error):
    traceback.print_exc()
    response = flask.jsonify({"error": str(error)})
    response.status_code = 400
    return response


@app.route('/api/v1/duts', methods=['GET'])
@app.route('/api/v1/machines', methods=['GET'])   # Deprecated
def get_machine_list():
    with app.app_context():
        mars = flask.current_app.mars

    return {
        "duts": dict([(m.id, m) for m in mars.known_machines])
    }


@app.route('/api/v1/dut/', methods=['POST', 'PUT'])
@app.route('/api/v1/machine/', methods=['POST', 'PUT'])   # Deprecated
def machine_add_or_update():
    with app.app_context():
        mars = flask.current_app.mars

    data = flask.request.json

    for key in data:
        if key not in {"id", "base_name", "tags", "manual_tags", "mac_address",
                       "ip_address", "local_tty_device", "ready_for_service",
                       "is_retired"}:
            raise ValueError(f"The field {key} cannot be set/modified")

    # Be backwards compatible with the older machine registration container that
    # did not specify the machine ID
    if 'id' not in data and data.get('mac_address'):
        data['id'] = data['mac_address']

    machine = mars.machine_discovered(data, update_if_already_exists=True)
    return CustomJSONEncoder().default(machine)


@app.route('/api/v1/dut/<machine_id>/', methods=['GET'])
@app.route('/api/v1/dut/<machine_id>', methods=['GET'])
@app.route('/api/v1/machine/<machine_id>/', methods=['GET'])   # Deprecated
@app.route('/api/v1/machine/<machine_id>', methods=['GET'])   # Deprecated
def machine_detail_get(machine_id):
    with app.app_context():
        mars = flask.current_app.mars

    machine = mars.get_machine_by_id(machine_id, raise_if_missing=True)
    return CustomJSONEncoder().default(machine)


@app.route('/api/v1/dut/<machine_id>/quick_check', methods=['GET', 'POST'])
@app.route('/api/v1/machine/<machine_id>/quick_check', methods=['GET', 'POST'])   # Deprecated
def machine_quick_check(machine_id):
    with app.app_context():
        mars = flask.current_app.mars

    machine = mars.get_machine_by_id(machine_id, raise_if_missing=True)

    if flask.request.method in ['GET']:
        ret = "true\n" if machine.quick_check_queued.is_set() else "false\n"
        return flask.make_response(ret, 200)
    elif flask.request.method in ['POST']:
        machine.quick_check_queued.set()
        return flask.make_response(f"Quick check queued for machine {machine.full_name}\n", 200)


@app.route('/api/v1/dut/<machine_id>', methods=['DELETE'])
@app.route('/api/v1/machine/<machine_id>', methods=['DELETE'])   # Deprecated
def machine_remove(machine_id):
    with app.app_context():
        mars = flask.current_app.mars

    machine = mars.get_machine_by_id(machine_id, raise_if_missing=True)

    if mars.remove_machine(machine_id):
        return flask.make_response(f"The machine {machine.full_name} was removed\n", 200)
    else:
        return flask.make_response(f"The machine {machine.full_name} does not exist\n", 404)


@app.route('/api/v1/dut/<machine_id>', methods=['PATCH'])
@app.route('/api/v1/machine/<machine_id>', methods=['PATCH'])   # Deprecated
def machine_update(machine_id):
    data = flask.request.get_json()

    for key in data:
        if key not in {"pdu_off_delay", "comment", "ready_for_service", "is_retired", "manual_tags"}:
            raise ValueError(f"The field {key} is invalid")

    with app.app_context():
        mars = flask.current_app.mars

    machine = mars.get_machine_by_id(machine_id, raise_if_missing=True)

    if pdu_off_delay := data.get("pdu_off_delay"):
        try:
            data["pdu_off_delay"] = float(pdu_off_delay)
        except ValueError:
            raise ValueError("Please provide the delay using a float or integer")

    machine.update_fields(data)
    return flask.make_response(f"Updated {machine.full_name}: {data}\n", 200)


@app.route('/boot/<machine_id>/boot.ipxe', methods=['GET'])
def machine_ipxe_boot_script(machine_id):
    with app.app_context():
        mars = flask.current_app.mars
        boots = flask.current_app.boots

    machine = mars.get_machine_by_id(machine_id)
    args = flask.request.args
    return boots.ipxe_boot_script(machine, platform=args.get("platform"), buildarch=args.get("buildarch"))


@app.route('/api/v1/pdus', methods=['GET'])
def get_pdus_list():
    pdus = {}

    with app.app_context():
        mars = flask.current_app.mars

    with mars.db as mars_db:
        for name in mars_db.pdus:
            pdus[name] = mars.get_pdu_by_name(name)

    return {
        "pdus": pdus
    }


@app.route('/api/v1/pdu/<pdu_name>', methods=['GET'])
def get_pdu(pdu_name):
    pdu = find_pdu(pdu_name)
    return flask.jsonify(pdu)


@app.route('/api/v1/pdu/<pdu_name>/port/<port_id>', methods=['GET', 'PATCH'])
def get_pdu_port(pdu_name, port_id):

    if flask.request.method in ['GET']:
        port = find_pdu_port(pdu_name, port_id)
        return flask.jsonify(port)

    elif flask.request.method in ['PATCH']:
        pdu_port = find_pdu_port(pdu_name, port_id)
        pdu = find_pdu(pdu_name)

        data = flask.request.get_json()

        for key in data:
            if key not in {"state", "reserved"}:
                raise ValueError(f"The field {key} is invalid")

        if (pdu_port.reserved and "reserved" not in data):
            raise ValueError(f'The port {port_id} on PDU "{pdu_name}" is reserved!')

        with app.app_context():
            mars = flask.current_app.mars
        with mars.db as mars_db:
            for machine in mars_db.duts.values():
                # port_id is a string but user could enter an integer and we can make it work
                if machine.pdu == pdu_name and str(machine.pdu_port_id) == str(port_id):
                    # If the machine is retired, it's OK to change the state
                    if (machine.is_retired and "state" in data):
                        break
                    raise ValueError(f'It\'s not possible to modify port {port_id} on PDU "{pdu_name}", '
                                     f'because it\'s already in use by machine {machine.full_name}.')

            if state := data.get("state"):
                if state.upper() in ["ON", "OFF"]:
                    new_state = PDUState[state.upper()]
                    pdu_port.set(new_state)
                    return flask.make_response(f'Turning {state} port {port_id} on PDU "{pdu_name}"\n', 200)
                else:
                    raise ValueError("Invalid state set, valid states are ON and OFF.")

            if state := data.get("reserved"):
                if state in ["false", "False"]:
                    pdu.unreserve_port(port_id)
                    return flask.make_response(f'Un-reserved port {port_id} on PDU "{pdu_name}"\n', 200)
                elif state in ["False", "True"]:
                    pdu.reserve_port(port_id)
                    return flask.make_response(f'Reserved port {port_id} on PDU "{pdu_name}"\n', 200)
                else:
                    raise ValueError("Invalid value to reserve a port, valid states are True and False.")


# When the request is made with a POST including a PDU and a port_id,
# it'll start a discovery process powering and updating the mars.discover_data
# with this information
# For request with a GET, it'll give the information in mars.discover_data
# so the user can see if there is a discovery process ongoing and since when.
# Finally if the method used is DELETE, discover_data will be deleted
@app.route('/api/v1/dut/discover', methods=['POST', 'GET', 'DELETE'])
@app.route('/api/v1/machine/discover', methods=['POST', 'GET', 'DELETE'])   # Deprecated
def discover_machine():
    with app.app_context():
        mars = flask.current_app.mars

    # show if there is a discovery in progress
    if flask.request.method in ['GET']:
        return flask.jsonify(mars.discover_data)

    # discover_data will be erased
    if flask.request.method in ['DELETE']:
        with mars.db as mars_db:
            if mars.discover_data:
                dd_pdu = mars.discover_data.get('pdu')
                dd_port_id = mars.discover_data.get('port_id')
                pdu_port = find_pdu_port(dd_pdu, dd_port_id)
                mars.discover_data = {}

                if pdu_port.reserved:
                    raise ValueError(f'Port {dd_port_id} on PDU "{dd_pdu}" is reserved!')

                # Be 100% sure we're not stopping any device!
                for machine in mars_db.duts.values():
                    # dd_port_id is a string but user could enter an integer and we can make it work
                    if machine.pdu == dd_pdu and str(machine.pdu_port_id) == str(dd_port_id):
                        raise ValueError(f'Port {dd_port_id} on PDU "{dd_pdu}" is already assigned.')

                pdu_port.set(PDUState.OFF)

                return flask.make_response("Discovery process canceled and port turned OFF.", 200)
            else:
                raise ValueError("There are no discovery processes in progress.")

    # The following code is only run when method is POST
    if mars.discover_data:
        raise ValueError("There is a discovery process running already.")

    data = flask.request.get_json()

    for key in data:
        if key not in {"pdu", "port_id", "timeout"}:
            raise ValueError(f"The field {key} is invalid")

    if not all(['port_id' in data, 'pdu' in data]):
        raise ValueError("You're missing at least one of the two required fields: pdu and port_id")

    dd_pdu = data.get('pdu')
    dd_port_id = data.get('port_id')

    with mars.db as mars_db:
        pdu_port = find_pdu_port(dd_pdu, dd_port_id)

        for machine in mars_db.duts.values():
            # port_id is a string but user could enter an integer and we can make it work
            if machine.pdu == dd_pdu and str(machine.pdu_port_id) == str(dd_port_id):
                raise ValueError(f'Port {dd_port_id} on PDU "{dd_pdu}" is already assigned.')

    if pdu_port.state == PDUState.ON:
        raise ValueError(f'Port {dd_port_id} on PDU "{dd_pdu} is already in use!')

    if pdu_port.reserved:
        raise ValueError(f'Port {dd_port_id} on PDU "{dd_pdu} is reserved!')

    # Launch discovery... the machine behind the PDU port should start
    pdu_port.set(PDUState.ON)

    if pdu_port.state == PDUState.ON:
        if data.get('timeout'):
            sec = int(data.get('timeout'))
        else:
            sec = 150

        mars.discover_data = {
            "pdu": dd_pdu,
            "port_id": dd_port_id,
            "date": datetime.now(),
            "timeout": int(sec),
            "started_at": time.monotonic(),
        }

        return flask.make_response(f'Booting machine behind port {dd_port_id} on PDU "{dd_pdu}".'
                                   f'Discovery will time out after {sec} seconds.\n', 200)
    else:
        raise ValueError(f'Failed to turn ON the port {dd_port_id} on PDU "{dd_pdu}"')


@app.route('/api/v1/dut/<machine_id>/cancel_job', methods=['POST'])
@app.route('/api/v1/machine/<machine_id>/cancel_job', methods=['POST'])   # Deprecated
def cancel_job_machine(machine_id):

    with app.app_context():
        mars = flask.current_app.mars

    m = mars.get_machine_by_id(machine_id, raise_if_missing=True)

    if m.state != DUTState.RUNNING:
        raise ValueError(f"Machine {m.full_name} isn't running a job. "
                         f"Current state is {m.state.name}")
    else:
        m.cancel_job()
        return flask.make_response(f"Canceling current job in machine {m.full_name}\n", 200)


@app.route('/api/v1/full-state', methods=['GET'])
def full_state():
    fs = {}

    with app.app_context():
        mars = flask.current_app.mars

    # Collect from the MarsDB all the information we need to generate the full state, then free the lock
    with mars.db as mars_db:
        fs['pdus'] = {p: mars.get_pdu_by_name(p) for p in mars_db.pdus}
        fs['duts'] = {m.id: m for m in mars.known_machines}
        fs['discover'] = mars.discover_data

    # Add a a field to show what is the current time on the server
    fs['now'] = datetime.now()

    return flask.jsonify(fs)


@app.route('/api/v1/jobs', methods=['POST'])
def post_job():
    def find_suitable_machine(target):
        with app.app_context():
            mars = flask.current_app.mars

        wanted_tags = set(target.tags)

        # If the target id is specified, check the tags
        if target.id is not None:
            machine = mars.get_machine_by_id(target.id)
            if machine is None:
                return None, 404, f"Unknown machine with ID {target.id}"
            elif not wanted_tags.issubset(machine.all_tags):
                return None, 406, (f"The machine {target.id} does not matching tags "
                                   f"(asked: {wanted_tags}, actual: {machine.all_tags})")
            elif machine.state != DUTState.IDLE:
                return None, 409, (f"The machine {target.id} is unavailable: "
                                   f"Current state is {machine.state.name}")
            elif machine.is_retired:
                return None, 409, (f"The machine {target.id} is retired.")
            return machine, 200, None
        else:
            found_a_candidate_machine = False
            for machine in mars.known_machines:
                if not wanted_tags.issubset(machine.all_tags):
                    continue
                if machine.is_retired:
                    continue

                found_a_candidate_machine = True
                if machine.state == DUTState.IDLE:
                    return machine, 200, "success"

            if found_a_candidate_machine:
                return None, 409, f"All machines matching the tags {wanted_tags} are busy"
            else:
                return None, 406, f"No active machines found matching the tags {wanted_tags}."

    def check_minio_credentials(job_request):
        credentials = job_request.minio_credentials

        # If no groups are requested, then exit directly
        if job_request.minio_groups is None or len(job_request.minio_groups) == 0:
            return True, ""

        # Some groups are requested, make sure some credentials have been set
        if credentials is None:
            return False, "Requested access to some groups, but the credentials are missing"

        # Make sure all the requested groups are in the list of groups the
        # provided-credentials have access to
        try:
            timestamp = int(datetime.now().timestamp())
            client = MinioClient(user=credentials.access_key,
                                 secret_key=credentials.secret_key,
                                 alias=f"a_{job_request.job_id}-{timestamp}")

            user_groups = set(client.groups_user_is_in())
            for group in job_request.minio_groups:
                if group not in user_groups:
                    return False, (f"The provided MinIO credentials do not belong to the group {group}")

            return True, ""
        except ValueError as e:
            return False, str(e)
        finally:
            try:
                client.remove_alias()
            except UnboundLocalError:
                pass

    # Grab the post_job lock, to prevent allocating two jobs for the same machine
    with app.app_context():
        with flask.current_app.post_job_lock:
            try:
                parsed = JobRequest.parse(flask.request)
            except InvalidTarballFile:
                return flask.make_response("Invalid tarball file", 420)

            try:
                ok, error_msg = check_minio_credentials(parsed)
                if ok:
                    machine, error_code, error_msg = find_suitable_machine(parsed.target)
                    if machine is not None:
                        if machine.state != DUTState.IDLE:
                            raise ValueError(f"The machine isn't idle: Current state is {machine.state.name}")

                        try:
                            machine.start_job(parsed)
                        except Exception as e:
                            error_code = 500
                            error_msg = str(e)
                else:
                    error_code = 403

                if parsed.version == 0:
                    response = {
                        "reason": error_msg
                    }
                elif parsed.version == 1:
                    response = {
                        # protocol version
                        "version": 1,
                        "error_msg": error_msg

                        # TODO: Store the job in memory, and show the ID here
                    }
                return flask.make_response(flask.jsonify(response), error_code)
            finally:
                # Make sure not to leak any ressource
                parsed.cleanup()


def sigterm_handler(_signo, _stack_frame):
    # Raises SystemExit(0):
    sys.exit(0)


def run():  # pragma: nocover
    # Make sure the farm name has been set
    if config.FARM_NAME is None:
        raise ValueError("Please set the FARM_NAME environment variable")

    # Check the Minio credentials: This will set up an alias, which fails if
    # the credentials are invalid and a ValueError will be raised, thus exiting
    # the process.
    MinioClient()

    pdu_daemon = PDUDaemon()

    # Create all the workers based on the machines found in MaRS
    mars = Mars(pdu_daemon)
    mars.start()

    # Start the network boot service
    # FIXME: Tidy this stuff up! Just give a boots root and descend from there.
    boots = BootService(mars=mars, config_paths={
        'TFTP_DIR': config.BOOTS_TFTP_ROOT,
    })

    # Start flask
    try:
        signal.signal(signal.SIGTERM, sigterm_handler)

        with app.app_context():
            flask.current_app.mars = mars
            flask.current_app.boots = boots
            flask.current_app.post_job_lock = multiprocessing.Lock()

        common_flags = {
            'asyncore_use_poll': True,
        }

        if http_sockets := get_sockets_by_name(config.EXECUTOR_HTTP_IPv4_SOCKET_NAME,
                                               socket.AF_INET, socket.SOCK_STREAM):
            serve(app, sockets=http_sockets, **common_flags)
        else:
            serve(app, host=config.EXECUTOR_HOST, port=config.EXECUTOR_PORT, **common_flags)
    finally:
        # Shutdown
        mars.stop(wait=True)
        pdu_daemon.stop()


if __name__ == '__main__':  # pragma: nocover
    run()



from functools import cached_property

from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
import requests

from .. import PDU, PDUPort, PDUState


class TasmotaPDU(PDU):
    @property
    def requests_retry_session(self, retries=3, backoff_factor=1,
                               status_forcelist=[], session=None):  # pragma: nocover
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def url(self, path):
        hostname = self.config['hostname']
        return f"http://{hostname}{path}"

    def get(self, path):
        r = self.requests_retry_session.get(self.url(path))
        r.raise_for_status()
        return r.json()

    @cached_property
    def num_ports(self):
        # NOTE: We really want to have a stable amount of ports, even in the presence
        # of network issues, so let's depend on the one thing we have: the status

        status = self.raw_status.get('Status', {})

        friendly_names = status.get("FriendlyName")
        if type(friendly_names) is list:
            port_count = len(friendly_names)
        else:  # pragma: nocover
            port_count = 1

        return port_count

    def __init__(self, name, config, reserved_port_ids=[]):
        super().__init__(name, config, reserved_port_ids)

        self.raw_status = self.get('/cm?cmnd=Status')

        self._ports = [PDUPort(self, str(p)) for p in range(self.num_ports)]

    @property
    def ports(self):
        return self._ports

    def __Power_to_PDUState(self, power):  # pragma: nocover
        if power in ["ON", "OFF"]:
            return PDUState[power]
        elif power.lower() in ['1', 'true']:
            return PDUState.ON
        elif power.lower() in ['0', 'false']:
            return PDUState.OFF
        else:
            raise ValueError(f"Unknown power status '{power}'")

    def __port_id_to_power_id(self, port_id):
        return int(port_id) + 1

    def _get_port_state(self, poid, raw_port):
        state = raw_port.get('POWER')
        if not state:
            state = raw_port.get(f'POWER{poid}')
        return self.__Power_to_PDUState(state)

    def set_port_state(self, port_id, state):
        if state in [PDUState.ON, PDUState.OFF]:
            poid = self.__port_id_to_power_id(port_id)
            raw_port = self.get(f'/cm?cmnd=Power{poid}%20{state.name}')
            return self._get_port_state(poid, raw_port) == state

    def get_port_state(self, port_id):
        poid = self.__port_id_to_power_id(port_id)
        raw_port = self.get(f'/cm?cmnd=Power{poid}')
        return self._get_port_state(poid, raw_port)



from unittest.mock import MagicMock, patch
import pytest

from server.pdu import PDUState, PDUPort
from server.pdu.drivers.shelly import ShellyPDU


class ShellyPlugMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "type": "SHPLG2-1",
                "mac": "C45BBE49EAA7",
                "auth": False,
                "fw": "20221027-102248/v1.12.1-ga9117d3",
                "discoverable": True,
                "num_outputs": 1,
                "num_meters": 1
            }
        elif url == f"{url_prefix}/settings/relay/0":
            ret = {
                "name": None,
                "appliance_type": "Computer",
                "ison": True,
                "has_timer": False,
                "default_state": "last",
                "auto_on": 0,
                "auto_off": 0,
                "schedule": False,
                "schedule_rules": [],
                "max_power": 3500
            }
        elif url == f"{url_prefix}/relay/0":
            ret = {
                "ison": True,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "overpower": False,
                "source": "input"
            }
        elif url == f"{url_prefix}/relay/0?turn=off":
            ret = {
                "ison": False,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "overpower": False,
                "source": "http"
            }
        # elif url == f"{url_prefix}/meter/0":
        #     ret = {
        #         "power": 218.11,
        #         "overpower": 0,
        #         "is_valid": True,
        #         "timestamp": 1684923696,
        #         "counters": [
        #             193.84,
        #             195.551,
        #             201.603
        #         ],
        #         "total": 6164469
        #     }
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', ShellyPlugMock)
def test_shelly_plug():
    pdu = ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.gen == 1
    assert pdu.num_ports == 1

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label=None)]
    assert pdu.get_port_state("0") == PDUState.ON
    assert pdu.set_port_state("0", PDUState.OFF)


class ShellyPlugSMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "type": "SHPLG-S",
                "mac": "7C87CEB519DB",
                "auth": False,
                "fw": "20230503-101129/v1.13.0-g9aed950",
                "discoverable": True,
                "longid": 1,
                "num_outputs": 1,
                "num_meters": 1
            }
        elif url == f"{url_prefix}/settings/relay/0":
            ret = {
                "name": "My channel",
                "appliance_type": "General",
                "ison": True,
                "has_timer": False,
                "default_state": "off",
                "auto_on": 0.00,
                "auto_off": 0.00,
                "schedule": False,
                "schedule_rules": [],
                "max_power": 1800
            }
        elif url == f"{url_prefix}/relay/0":
            ret = {
                "ison": True,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "overpower": False,
                "source": "http",
            }
        elif url == f"{url_prefix}/relay/0?turn=off":
            ret = {
                "ison": False,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "overpower": False,
                "source": "http",
            }
        # elif url == f"{url_prefix}/meter/0":
        #     ret = {
        #         "power": 95.55,
        #         "overpower": 0,
        #         "is_valid": True,
        #         "timestamp": 1684923780,
        #         "counters": [
        #             107.304,
        #             90.008,
        #             86.23
        #         ],
        #         "total": 2893636
        #     }
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', ShellyPlugSMock)
def test_shelly_plug_s():
    pdu = ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.gen == 1
    assert pdu.num_ports == 1

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label="My channel")]
    assert pdu.get_port_state("0") == PDUState.ON
    assert pdu.set_port_state("0", PDUState.OFF)


class Shelly1LMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "type": "SHSW-L",
                "mac": "98CDAC1EB380",
                "auth": False,
                "fw": "20221027-091704/v1.12.1-ga9117d3",
                "discoverable": True,
                "longid": 1,
                "num_inputs": 2,
                "num_outputs": 1,
                "num_meters": 1
            }
        elif url == f"{url_prefix}/settings/relay/0":
            ret = {
                "name": None,
                "appliance_type": "General",
                "ison": True,
                "has_timer": False,
                "default_state": "last",
                "btn1_type": "edge",
                "btn1_reverse": 0,
                "btn2_type": "toggle",
                "btn2_reverse": 0,
                "swap_inputs": False,
                "auto_on": 0,
                "auto_off": 0,
                "schedule": False,
                "schedule_rules": []
            }
        elif url == f"{url_prefix}/relay/0":
            ret = {
                "ison": True,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "source": "http"
            }
        elif url == f"{url_prefix}/relay/0?turn=off":
            ret = {
                "ison": False,
                "has_timer": False,
                "timer_started": 0,
                "timer_duration": 0,
                "timer_remaining": 0,
                "source": "http"
            }
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', Shelly1LMock)
def test_shelly_1L():
    pdu = ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.gen == 1
    assert pdu.num_ports == 1

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label=None)]
    assert pdu.get_port_state("0") == PDUState.ON
    assert pdu.set_port_state("0", PDUState.OFF)


class ShellyPlus2PMMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "name": "My ShellyPlus2PM",
                "id": "shellyplus2pm-c049ef860c58",
                "mac": "C049EF860C58",
                "model": "SNSW-102P16EU",
                "gen": 2,
                "fw_id": "20220527-091739/0.10.2-beta4-gecc3a61",
                "ver": "0.10.2-beta4",
                "app": "Plus2PM",
                "auth_en": False,
                "auth_domain": None,
                "profile": "switch"
            }
        elif url == f"{url_prefix}/rpc/Switch.GetConfig?id=0":
            ret = {
                "id": 0,
                "name": "Channel1",
                "in_mode": "follow",
                "initial_state": "on",
                "auto_on": False,
                "auto_on_delay": 60.00,
                "auto_off": False,
                "auto_off_delay": 60.00,
                "power_limit": 2800,
                "voltage_limit": 280,
                "current_limit": 10.000
            }
        elif url == f"{url_prefix}/rpc/Switch.GetConfig?id=1":
            ret = {
                "id": 1,
                "name": "Channel2",
                "in_mode": "follow",
                "initial_state": "off",
                "auto_on": False,
                "auto_on_delay": 60.00,
                "auto_off": False,
                "auto_off_delay": 60.00,
                "power_limit": 2800,
                "voltage_limit": 280,
                "current_limit": 10.000
            }
        elif url == f"{url_prefix}/rpc/Switch.GetStatus?id=0":
            ret = {
                "id": 0,
                "source": "HTTP",
                "output": False,
                "apower": 0.0,
                "voltage": 235.6,
                "current": 0.000,
                "pf": 0.00,
                "aenergy": {
                    "total": 4120.351,
                    "by_minute": [0.000, 0.000, 0.000],
                    "minute_ts": 1684838501
                },
                "temperature": {
                    "tC": 27.7,
                    "tF": 81.9
                }
            }
        elif url == f"{url_prefix}/rpc/Switch.GetStatus?id=1":
            ret = {
                "id": 1,
                "source": "HTTP",
                "output": True,
                "apower": 0.0,
                "voltage": 240.6,
                "current": 0.000,
                "pf": 0.00,
                "aenergy": {
                    "total": 5120.351,
                    "by_minute": [0.000, 0.000, 0.000],
                    "minute_ts": 1684838501
                },
                "temperature": {
                    "tC": 27.7,
                    "tF": 81.9
                }
            }
        elif url == f"{url_prefix}/rpc/Switch.Set?id=0&on=true":
            ret = {"was_on": False}
        elif url == f"{url_prefix}/rpc/Switch.Set?id=1&on=false":
            ret = {"was_on": True}
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', ShellyPlus2PMMock)
def test_shelly_plus_2pm():
    pdu = ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'}, reserved_port_ids=['1'])

    assert pdu.gen == 2
    assert pdu.num_ports == 2

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label="Channel1"),
                         PDUPort(pdu=pdu, port_id="1", label="Channel2")]
    assert pdu.get_port_state("0") == PDUState.OFF
    assert pdu.get_port_state("1") == PDUState.ON

    assert pdu.set_port_state("0", PDUState.ON)
    assert pdu.set_port_state("1", PDUState.OFF)


class ShellyPlusPlugSMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "name": "My ShellyPlusPlugS",
                "id": "shellyplusplugs-80646fd61f2c",
                "mac": "80646FD61F2C",
                "model": "SNPL-00112EU",
                "gen": 2,
                "fw_id": "20230510-081027/0.14.4-g4da93ee",
                "ver": "0.14.4",
                "app": "PlusPlugS",
                "auth_en": False,
                "auth_domain": None,
            }
        elif url == f"{url_prefix}/rpc/Switch.GetConfig?id=0":
            ret = {
                "id": 0,
                "name": "Channel",
                "initial_state": "off",
                "auto_on": False,
                "auto_on_delay": 60,
                "auto_off": False,
                "auto_off_delay": 60,
                "power_limit": 2500,
                "voltage_limit": 280,
                "current_limit": 12
            }
        elif url == f"{url_prefix}/rpc/Switch.GetStatus?id=0":
            ret = {
                "id": 0,
                "source": "init",
                "output": False,
                "apower": 0,
                "voltage": 0,
                "current": 0,
                "aenergy": {
                    "total": 0,
                    "by_minute": [0, 0, 0],
                    "minute_ts": 1684910834
                },
                "temperature": {
                    "tC": 36.3,
                    "tF": 97.4
                }
            }
        elif url == f"{url_prefix}/rpc/Switch.Set?id=0&on=true":
            ret = {"was_on": False}
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', ShellyPlusPlugSMock)
def test_shelly_plus_plug_s():
    pdu = ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.gen == 2
    assert pdu.num_ports == 1

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label="Channel")]
    assert pdu.get_port_state("0") == PDUState.OFF
    assert pdu.set_port_state("0", PDUState.ON)


class ShellyFutureDeviceMock:
    def get(url, timeout=None):
        assert timeout == 1

        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/shelly":
            ret = {
                "name": "My ShellyPlusPlus2PM",
                "id": "shellyplusplus2pm-c049ef860c58",
                "mac": "C049EF860C58",
                "model": "SNSW-102P16EU",
                "gen": 2,
                "fw_id": "20220527-091739/0.10.2-beta4-gecc3a61",
                "ver": "0.10.2-beta4",
                "app": "PlusPlus2PM",
                "auth_en": False,
                "auth_domain": None,
                "profile": "switch"
            }
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.shelly.ShellyPDU.requests_retry_session', ShellyFutureDeviceMock)
def test_shelly_future_device():
    with pytest.raises(ValueError) as excinfo:
        ShellyPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert "Unknown Shelly device: gen=2, type=None, model=SNSW-102P16EU, app=PlusPlus2PM" in str(excinfo.value)



from unittest.mock import patch
import pytest

from server.pdu import PDUState
from server.pdu.drivers.kernelchip import KernelChipPDU


class KernelChipBaseMock:
    def __init__(self, *args, **kwargs):
        self.password = kwargs.get('password', None)
        self.states = b'1111111011101101110010101111'
        self._data_to_recv = b'#FLG,AB,11,11\r\nJConfig from FLASH\r\n'
        self.auth = self.password is None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        (ret, self._data_to_recv) = (self._data_to_recv[0:2], self._data_to_recv[2:])
        return ret

    def sendall(self, data):
        if not data.endswith(b'\r\n'):  # pragma: no cover
            self._data_to_recv += b'#ERR\r\n'
            return

        (ke, *msg) = data[:-2].split(b',')
        if ke != b'$KE':  # pragma: no cover
            self._data_to_recv += b'#ERR\r\n'
        elif msg == []:
            self._data_to_recv += b'#OK\r\n'
        elif msg == [b'INF']:  # pragma: no cover
            self._data_to_recv += b'#INF,Laurent-XXX,XXXX,XXXX-XXXX-XXXX-XXXX\r\n'
        elif len(msg) == 3 and msg[0] == b'PSW' and msg[1] == b'SET':
            if self.password == msg[2]:
                self._data_to_recv += b'#PSW,SET,OK\r\n'
                self.auth = True
            else:
                self._data_to_recv += b'#PSW,SET,ERR\r\n'
        elif msg[0] == b'REL':
            # no cover, because the driver verifies password earlier
            if not self.auth:  # pragma: no cover
                self._data_to_recv += b'#Access denied. Password is needed.\r\n'
                return

            port = int(str(msg[1], 'ascii'))
            if port < 1 or port > len(self.states):  # pragma: no cover
                self._data_to_recv += b'#REL,ERR\r\n'
            else:
                self.states = self.states[:port - 1] + msg[2][:] + self.states[port:]
                self._data_to_recv += b'#REL,OK\r\n'
        elif msg[0] == b'RDR' and len(msg) == 2:
            if not self.auth:
                self._data_to_recv += b'#Access denied. Password is needed.\r\n'
                return

            if msg[1] == b'ALL':
                self._data_to_recv += b'#RDR,ALL,' + self.states + b'\r\n'
            else:  # pragma: no cover
                port = int(str(msg[1], 'ascii'))
                if port < 1 or port > len(self.states):
                    self._data_to_recv += b'#RDR,ERR\r\n'
                elif self.states[port - 1] == ord(b'1'):
                    self._data_to_recv += b'#RDR,' + msg[1] + b',1\r\n'
                else:
                    self._data_to_recv += b'#RDR,' + msg[1] + b',0\r\n'
        else:  # pragma: no cover
            self._data_to_recv += b'#ERR\r\n'

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=KernelChipBaseMock)
def test_kernelchip_no_password_successful_connection():
    pdu = KernelChipPDU("TestPDU", {})

    assert len(pdu.ports) == 28
    assert pdu.get_port_state(1) == PDUState.ON
    assert pdu.get_port_state(2) == PDUState.ON
    assert pdu.get_port_state(3) == PDUState.ON
    assert pdu.get_port_state(4) == PDUState.ON
    assert pdu.get_port_state(5) == PDUState.ON
    assert pdu.get_port_state(6) == PDUState.ON
    assert pdu.get_port_state(7) == PDUState.ON
    assert pdu.get_port_state(8) == PDUState.OFF
    assert pdu.get_port_state(9) == PDUState.ON
    assert pdu.get_port_state(10) == PDUState.ON
    assert pdu.get_port_state(11) == PDUState.ON
    assert pdu.get_port_state(12) == PDUState.OFF
    assert pdu.get_port_state(13) == PDUState.ON
    assert pdu.get_port_state(14) == PDUState.ON
    assert pdu.get_port_state(15) == PDUState.OFF
    assert pdu.get_port_state(16) == PDUState.ON
    assert pdu.get_port_state(17) == PDUState.ON
    assert pdu.get_port_state(18) == PDUState.ON
    assert pdu.get_port_state(19) == PDUState.OFF
    assert pdu.get_port_state(20) == PDUState.OFF
    assert pdu.get_port_state(21) == PDUState.ON
    assert pdu.get_port_state(22) == PDUState.OFF
    assert pdu.get_port_state(23) == PDUState.ON
    assert pdu.get_port_state(24) == PDUState.OFF
    assert pdu.get_port_state(25) == PDUState.ON
    assert pdu.get_port_state(26) == PDUState.ON
    assert pdu.get_port_state(27) == PDUState.ON
    assert pdu.get_port_state(28) == PDUState.ON


@patch('socket.socket', new=KernelChipBaseMock)
def test_kernelchip_get_port_state__invalid_port_id():
    pdu = KernelChipPDU("TestPDU", {})

    assert len(pdu.ports) == 28
    with pytest.raises(ValueError) as excinfo:
        pdu.get_port_state(29)

    assert str(excinfo.value) == "Invalid port id"


@patch('socket.socket', new=KernelChipBaseMock)
def test_kernelchip_set_port_state():
    pdu = KernelChipPDU("TestPDU", {})

    assert len(pdu.ports) == 28
    pdu.set_port_state(1, PDUState.OFF)


@patch('socket.socket', new=KernelChipBaseMock)
def test_kernelchip_set_port_state__bad_invalid_port_id():
    pdu = KernelChipPDU("TestPDU", {})

    assert len(pdu.ports) == 28
    with pytest.raises(ValueError) as excinfo:
        pdu.set_port_state(29, PDUState.OFF)

    assert str(excinfo.value) == "Invalid port id"


@patch('socket.socket', new=KernelChipBaseMock)
def test_kernelchip_port_state_management():
    import time
    pdu = KernelChipPDU("TestPDU", {})

    assert pdu.get_port_state(1) == PDUState.ON
    time.sleep(0.5)
    assert pdu.get_port_state(1) == PDUState.ON


class KernelChipPasswordMock(KernelChipBaseMock):
    def __init__(self, *args, **kwargs):
        super().__init__(self, args, password=b'PassWord')


@patch('socket.socket', new=KernelChipPasswordMock)
def test_kernelchip_correct_password():
    pdu = KernelChipPDU("TestPDU", {'password': 'PassWord'})

    assert len(pdu.ports) == 28
    assert pdu.get_port_state(1) == PDUState.ON


@patch('socket.socket', new=KernelChipPasswordMock)
def test_kernelchip_incorrect_password():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {'password': 'NoPassWord'})

    assert str(excinfo.value) == "Password Verification failure"


@patch('socket.socket', new=KernelChipPasswordMock)
def test_kernelchip_no_password():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {})

    assert str(excinfo.value) == "Password is required"


class KernelChipInvalidFirstByteMock(KernelChipBaseMock):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self._data_to_recv = b'abc\r\n'


@patch('socket.socket', new=KernelChipInvalidFirstByteMock)
def test_kernelchip_parse_error__invalid_first_byte():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {})

    assert str(excinfo.value) == "Invalid first byte"


class KernelChipDuplicateCRMock(KernelChipBaseMock):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self._data_to_recv = b'abc\r\r\n'


@patch('socket.socket', new=KernelChipDuplicateCRMock)
def test_kernelchip_parse_error__duplicate_cr():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {})

    assert str(excinfo.value) == "Duplicate '\\r'"


class KernelChipRespondErrMock(KernelChipBaseMock):
    def sendall(self, data):
        self._data_to_recv += b'#ERR\r\n'


@patch('socket.socket', new=KernelChipRespondErrMock)
def test_kernelchip_parse_error__respond_err():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {})

    assert str(excinfo.value) == "Protocol error"


class KernelChipIncorrectRdrAllResponseMock(KernelChipBaseMock):
    def sendall(self, data):
        if data == b'$KE,RDR,ALL\r\n':
            self._data_to_recv += b'#RDR,ABC,123\r\n'
        else:
            super().sendall(data)


@patch('socket.socket', new=KernelChipIncorrectRdrAllResponseMock)
def test_kernelchip_parse_error__incorrect_rdr_all_response():
    with pytest.raises(ValueError) as excinfo:
        KernelChipPDU("TestPDU", {})

    assert str(excinfo.value) == "Failure getting relays state"


class KernelChipIncorrectRelResponseMock(KernelChipBaseMock):
    def sendall(self, data):
        if data.startswith(b'$KE,REL,'):
            self._data_to_recv += b'#REL,ABC\r\n'
        else:
            super().sendall(data)


@patch('socket.socket', new=KernelChipIncorrectRelResponseMock)
def test_kernelchip_parse_error__incorrect_rel_response():
    pdu = KernelChipPDU("TestPDU", {})

    with pytest.raises(ValueError) as excinfo:
        pdu.set_port_state(1, PDUState.OFF)

    assert str(excinfo.value) == "Incorrect response"



from collections import defaultdict
from dataclasses import asdict, field
from datetime import datetime
from subprocess import Popen, PIPE, DEVNULL
from enum import Enum
from typing import Dict, List, Optional
from threading import Thread, Event
import subprocess
import traceback
import tempfile
import socket
import requests
import tarfile
import fcntl
import time
import json
import dataclasses
import yaml
import sys
import os

from pydantic import field_validator
from pydantic.dataclasses import dataclass
import requests_unixsocket

from .boots import BootConfig
from .job import Job
from .logger import logger
from .message import JobStatus, MessageType, Message
from .pdu import PDUState
from . import config


@dataclass
class CallbackEndpoint:
    host: str
    port: int


@dataclass
class MinIOCredentials:
    access_key: Optional[str] = None
    secret_key: Optional[str] = None


# WARNING: This class is duplicated from job.py until marshmallow gets replaced by pydantic dataclasses
@dataclass
class Target:
    id: Optional[str] = None
    tags: Optional[List[str]] = field(default_factory=list)

    @field_validator('tags')
    @classmethod
    def tags_must_be_a_list(cls, v):
        if v is None:
            return []
        else:
            return v


class DUTState(Enum):
    WAIT_FOR_CONFIG = 0
    IDLE = 1
    TRAINING = 2
    QUEUED = 3
    RUNNING = 4
    RETIRED = 5
    UNKNOWN = 6
    QUICK_CHECK = 7
    BORKED = 8


@dataclass(kw_only=True)
class JobRequest:
    version: int
    raw_job: str
    target: Target
    callback_endpoint: CallbackEndpoint
    minio_credentials: MinIOCredentials

    job_id: Optional[str]
    http_headers: Dict[str, str] = field(default_factory=dict)
    minio_groups: List[str] = field(default_factory=list)
    job_bucket_initial_state_tarball_file_fd: int = 0

    def cleanup(self):
        if self.job_bucket_initial_state_tarball_file_fd > 0:
            try:
                os.close(self.job_bucket_initial_state_tarball_file_fd)
            except Exception:
                logger.error(f"ERROR: Failed to close the initial state tarball file fd!\n{traceback.format_exc()}")
            finally:
                # Unconditionally mark the fd as freed, even if we failed to
                # close it because it is better to leak the fd than potentially
                # closing someone else's fd later down the line when this object
                # gets garbage collected.
                self.job_bucket_initial_state_tarball_file_fd = 0

    def __del__(self):
        self.cleanup()

    @field_validator("job_id")
    @classmethod
    def job_id_isnt_empty(cls, v):
        if not v:
            now = int(datetime.utcnow().timestamp())
            v = f"untitled-{now}"
        return v

    @field_validator("callback_endpoint")
    @classmethod
    def callback_endpoint_is_set(cls, v):
        if v.host is None:
            raise ValueError("callback's host cannot be None. Leave empty to get the default value")
        if v.port is None:
            raise ValueError("callback's port cannot be None")
        return v

    @classmethod
    def parse(cls, request):
        if request.mimetype == "application/json":
            return JSONJobRequest(request)
        elif request.mimetype == "multipart/form-data":
            return MultipartJobRequest(request)
        else:
            raise ValueError("Unknown job request format")


# DEPRECATED: To be removed when we are sure all the clients out there have been updated
class JSONJobRequest(JobRequest):
    def __init__(self, request):
        job_params = request.json
        metadata = job_params["metadata"]
        job = Job.render_with_resources(job_params["job"])

        # Use the client-provided host callback if available, or default to the remote addr
        remote_addr = metadata.get("callback_host", request.remote_addr)
        endpoint = CallbackEndpoint(remote_addr, metadata.get("callback_port"))

        super().__init__(http_headers=dict(request.headers), version=0, raw_job=job_params["job"],
                         target=job.target, callback_endpoint=endpoint,
                         minio_credentials=MinIOCredentials())


class InvalidTarballFile(Exception):
    pass


class MultipartJobRequest(JobRequest):
    def __init__(self, request):
        metadata_file = request.files.get('metadata')
        if metadata_file is None:
            raise ValueError("No metadata file found")

        if metadata_file.mimetype != "application/json":
            raise ValueError("The metadata file has the wrong mimetype: "
                             "{metadata_file.mimetype}} instead of application/json")

        try:
            metadata = json.loads(metadata_file.read())
        except json.JSONDecodeError as e:
            raise ValueError(f"The metadata file is not a valid JSON file: {e.msg}")

        version = metadata.get('version')
        if version == 1:
            self.parse_v1(request, metadata)
        else:
            raise ValueError(f"Invalid request version {version}")

    def parse_v1(self, request, metadata):
        # Get the job file, and check its mimetype
        job_file = request.files['job']
        if job_file.mimetype != "application/x-yaml":
            raise ValueError("The metadata file has the wrong mimetype: "
                             "{job_file.mimetype}} instead of application/x-yaml")

        # Get the initial state tarball.
        # Since flask files can be either in memory or on the disk, let's write its content
        # to a temporary file, keep a reference open by tying it to the lifetime of this object,
        # then store the fd in job_bucket_initial_state_tarball_file_fd. This file descriptor
        # will then be passed to the executor instance that will execute the job at which point
        # the object can be deleted automatically and the filedescriptor will get closed.
        initial_state_tarball_file = request.files.get('job_bucket_initial_state_tarball_file', None)
        if initial_state_tarball_file and initial_state_tarball_file.mimetype != "application/octet-stream":
            raise ValueError("The job_bucket_initial_state_tarball file has the wrong mimetype: "
                             "{initial_state_tarball_file.mimetype}} instead of application/octet-stream")
        if initial_state_tarball_file:
            job_bucket_initial_state_tarball_file_fd, file_path = tempfile.mkstemp()

            # Make sure the file doesn't linger on the disk even if everything crashes
            os.unlink(file_path)

            with os.fdopen(job_bucket_initial_state_tarball_file_fd, 'w+b', closefd=False) as tmp_file:
                while chunk := initial_state_tarball_file.read(1000000):
                    tmp_file.write(chunk)

                # Check if the file passed is a valid tarball, or raise InvalidTarballFile
                # NOTE: We make sure to rewind the file before and after the check!
                tmp_file.seek(0)
                if not tarfile.is_tarfile(tmp_file):
                    raise InvalidTarballFile()
                tmp_file.seek(0)
        else:
            job_bucket_initial_state_tarball_file_fd = -1

        # Create a Job object
        raw_job = job_file.read().decode()
        job = Job.render_with_resources(raw_job)

        # Get the target that will run the job. Use the job's target by default,
        # but allow the client to override the target
        if "target" in metadata:
            target = metadata.get('target', {})
            job_target = Target(target.get('id'), target.get('tags', []))
        else:
            job_target = Target(id=job.target.id, tags=job.target.tags)

        # Use the client-provided host callback if available, or default to the remote addr
        callback = metadata.get('callback', {})
        remote_addr = callback.get("host", request.remote_addr)
        endpoint = CallbackEndpoint(remote_addr, callback.get("port"))

        # Parse the minio-related arguments request
        minio = metadata.get('minio', {})
        minio_credentials = minio.get('credentials', {})
        credentials = MinIOCredentials(access_key=minio_credentials.get("access_key"),
                                       secret_key=minio_credentials.get("secret_key"))

        super().__init__(http_headers=dict(request.headers), version=1, raw_job=raw_job,
                         target=job_target, callback_endpoint=endpoint,
                         job_bucket_initial_state_tarball_file_fd=job_bucket_initial_state_tarball_file_fd,
                         job_id=metadata.get('job_id'),
                         minio_credentials=credentials,
                         minio_groups=minio.get('groups', []))


class SergentHartmanState(Enum):
    IDLE = 0
    ENROLLING = 1
    QUICK_CHECK = 2
    REGISTRATION_FAILED = 3


class SergentHartman:
    def __init__(self, machine, boot_loop_counts=None, qualifying_rate=None, quick_check=None):
        super().__init__()

        if boot_loop_counts is None:
            boot_loop_counts = int(config.SERGENT_HARTMAN_BOOT_COUNT)

        if qualifying_rate is None:
            qualifying_rate = int(config.SERGENT_HARTMAN_QUALIFYING_BOOT_COUNT)

        if quick_check is None:
            quick_check = config.as_boolean("SERGENT_HARTMAN_QUICK_CHECK")

        self.machine = machine
        self.boot_loop_counts = boot_loop_counts
        self.qualifying_rate = qualifying_rate
        self.quick_check = quick_check
        self._state = None
        self.registration_failed_at = None

        # Tri-state boolean
        self.result = None

        self.set_state(SergentHartmanState.IDLE)

    @property
    def is_active(self):
        return self.state != SergentHartmanState.IDLE

    @property
    def state(self):
        def registration_failed_expiration_time():
            delay = int(config.SERGENT_HARTMAN_REGISTRATION_RETRIAL_DELAY)
            return self.registration_failed_at + delay if self.registration_failed_at is not None else 0

        if self._state == SergentHartmanState.ENROLLING and registration_failed_expiration_time() > time.monotonic():
            return SergentHartmanState.REGISTRATION_FAILED
        else:
            return self._state

    def set_state(self, state):
        def reset():
            self.is_machine_registered = False
            self.cur_loop = 0
            self.statuses = defaultdict(int)
            self.result = None
            self.registration_failed_at = None

        if self.state == state:
            return
        elif state == SergentHartmanState.IDLE:
            reset()
        elif self.state == SergentHartmanState.IDLE and state == SergentHartmanState.ENROLLING:
            pass
        elif self.state == SergentHartmanState.IDLE and state == SergentHartmanState.QUICK_CHECK:
            pass
        elif state == SergentHartmanState.REGISTRATION_FAILED:
            reset()
            self.registration_failed_at = time.monotonic()
            state = SergentHartmanState.ENROLLING
        else:
            raise ValueError(f"{self.state} -> {state} is an invalid transition")

        self._state = state

    def _next_task(self, callback_port):
        mid = self.machine.id

        if self.state == SergentHartmanState.ENROLLING:
            if not self.is_machine_registered:
                # Start by forcing the machine to register itself to make sure the
                # its configuration is up to date (especially the serial console
                # port). Loop until it succeeds!
                job_path = config.EXECUTOR_REGISTRATION_JOB

                logger.info("SergentHartman/%s - Try registering the machine", mid)
            else:
                # Check that we got the expected amount of reports
                if self.cur_loop != sum(self.statuses.values()):
                    raise ValueError("The previous next_task() call was not followed by a call to report()")

                # The registration went well, let's start the boot loop!
                self.cur_loop += 1

                statuses_str = [f"{status.name}: {values}" for status, values in self.statuses.items()]
                logger.info("SergentHartman/%s - loop %s/%s - statuses %s: "
                            "Execute one more round!",
                            mid,
                            self.cur_loop,
                            self.boot_loop_counts,
                            statuses_str)

                job_path = config.EXECUTOR_BOOTLOOP_JOB
        elif self.state == SergentHartmanState.QUICK_CHECK:
            logger.info("SergentHartman/%s - Initial check", mid)
            job_path = config.EXECUTOR_BOOTLOOP_JOB
        else:
            raise ValueError(f"There are no next tasks when the state is {self.state.name}")

        with open(job_path, "r") as f:
            raw_job = f.read()

            callback_endpoint = CallbackEndpoint(host="127.0.0.1", port=callback_port)
            return JobRequest(version=1, job_id=None, raw_job=raw_job,
                              target=Target(), callback_endpoint=callback_endpoint,
                              minio_credentials=MinIOCredentials())

    def _report(self, job_status, execution_time):
        mid = self.machine.id

        if self.state == SergentHartmanState.ENROLLING:
            if self.cur_loop == 0:
                if job_status != JobStatus.PASS:
                    delay = int(config.SERGENT_HARTMAN_REGISTRATION_RETRIAL_DELAY)
                    logger.warning((f"SergentHartman/{mid} - Registration failed with status {job_status.name}. "
                                    f"Retrying in {delay} second(s)"))
                    self.set_state(SergentHartmanState.REGISTRATION_FAILED)
                else:
                    self.is_machine_registered = True
                    if self.boot_loop_counts >= 1:
                        logger.info(f"SergentHartman/{mid} - Registration succeeded, moving on to the boot loop")
                    else:
                        logger.info(f"SergentHartman/{mid} - Registration succeeded, boot loops disabled")
                        self.result = True
            else:
                # We are in the boot loop
                self.statuses[job_status] += 1

                if self.cur_loop >= self.boot_loop_counts:
                    self.result = self.statuses[JobStatus.PASS] >= self.qualifying_rate
        elif self.state == SergentHartmanState.QUICK_CHECK:
            self.result = (job_status == JobStatus.PASS)

            if job_status == JobStatus.PASS:
                logger.info(f"SergentHartman/{mid} - Initial check successful after {execution_time:.2f} seconds")
            else:
                logger.error(f"SergentHartman/{mid} - Initial check failed after {execution_time:.2f} seconds")
        else:
            raise ValueError(f"Reporting unsupported on the state {self.state.name}")

    def execute_next_task(self, stop_event):
        if self.state in [SergentHartmanState.IDLE, SergentHartmanState.REGISTRATION_FAILED]:
            # Nothing to do here
            return

        if stop_event.is_set():
            return

        # If we are asked for fewer than 0 bootloops, consider Sergent Hartman disabled
        if self.boot_loop_counts < 0:
            self.result = True
            return

        # If we were asked to quick check but quick check was disabled, pretend we ran it
        # already
        if self.state == SergentHartmanState.QUICK_CHECK and not self.quick_check:
            self.result = True
            return

        start_time = time.monotonic()
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as tcp_server:
            tcp_server.bind(('', 0))
            tcp_server.listen(1)
            local_port = tcp_server.getsockname()[1]

            # Queue the job
            status = DUTState.UNKNOWN
            try:
                proc = self.machine.start_job(self._next_task(callback_port=local_port), quiet=True)

                print(f"Waiting for the executor to connect to our local port {local_port}")
                tcp_server.settimeout(5)
                try:
                    job_socket, _ = tcp_server.accept()
                except socket.timeout:
                    proc.kill()
                    proc.wait(timeout=10)
                    raise ValueError("The server failed to initiate a connection")

                # Set the resulting socket's timeout to blocking
                job_socket.settimeout(None)

                # Wait for the end message
                status_msg = JobStatus.UNKNOWN
                try:
                    while not stop_event.is_set():
                        msg = Message.next_message(job_socket)

                        if msg.msg_type == MessageType.SESSION_END:
                            status_msg = msg.status
                            break
                except Exception:
                    traceback.print_exc()

                # Close the socket to signal we are done
                job_socket.shutdown(socket.SHUT_RDWR)
                job_socket.close()

                # Wait for the process to die
                status_exit_code = JobStatus.UNKNOWN
                try:
                    proc.wait(timeout=10)
                    status_exit_code = JobStatus(proc.returncode)
                except subprocess.TimeoutExpired:
                    proc.kill()
                    raise ValueError("The process never finished")
                except ValueError:
                    logger.error(f"The exit status {proc.returncode} is an invalid value for JobStatus")

                # Check that the status reported by the SESSION_END message matches
                # the one we got from
                if status_msg != JobStatus.UNKNOWN and \
                   status_exit_code != JobStatus.UNKNOWN and \
                   status_msg != status_exit_code:
                    raise ValueError("Mismatch detected between the END_MESSAGE's and the exit code's status")

                # report the status
                status = status_msg if status_msg != JobStatus.UNKNOWN else status_exit_code
            finally:
                execution_time = time.monotonic() - start_time
                self._report(job_status=status, execution_time=execution_time)

    @property
    def is_available(self):
        return config.EXECUTOR_REGISTRATION_JOB or config.EXECUTOR_BOOTLOOP_JOB


def lock_fd(fd, attempts=100):
    last_exc = None

    for i in range(attempts):
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return
        except BlockingIOError as e:
            if e.errno == 11:
                last_exc = e
                time.sleep(0.01)
                continue
            else:
                raise e

    raise last_exc


@dataclass
class PduPortStats:
    last_polled: Optional[datetime] = None
    last_known_state: Optional[PDUState] = None
    last_shutdown: Optional[datetime] = None

    def serialize(self):
        fields = asdict(self)
        fields['last_known_state'] = fields['last_known_state'].value
        return fields

    @classmethod
    def from_pdu_port(cls, port):
        fields = {f.name: getattr(port, f.name) for f in dataclasses.fields(cls)}
        return cls(**fields)


class DUT(Thread):
    def __init__(self, mars, db_dut):
        self.mars = mars
        self.config_changed(db_dut)

        # Training / Qualifying process
        self.sergent_hartman = SergentHartman(self)

        # Queue a quick check
        self.quick_check_queued = Event()
        if self.ready_for_service and not self.is_retired:
            self.quick_check_queued.set()

        # Start the background thread that will manage the machine
        super().__init__(name=f'ExecutorThread-{self.id}')
        self.stop_event = Event()
        self.start()

    def stop_machine(self):
        self.stop_event.set()
        self.cancel_job()
        self.join()

    # Expose all the fields of the associated ConfigDUT object
    def __getattr__(self, attr):
        return getattr(self.db_dut, attr)

    @property
    def ready_for_service(self):
        return self.db_dut.ready_for_service

    @ready_for_service.setter
    def ready_for_service(self, val):
        self.update_fields({"ready_for_service": val})

    @property
    def is_retired(self):
        return self.db_dut.is_retired

    @is_retired.setter
    def is_retired(self, val):
        self.update_fields({"is_retired": val})

    @property
    def pdu_port(self):
        if self._pdu_port is None:
            self._pdu_port = self.mars.get_pdu_port_by_name(self.db_dut.pdu,
                                                            self.db_dut.pdu_port_id,
                                                            raise_if_missing=False,
                                                            timeout=0)
            if self._pdu_port:
                self._pdu_port.min_off_time = self.db_dut.pdu_off_delay

        return self._pdu_port

    def config_changed(self, db_dut=None):
        self.db_dut = db_dut

        # Invalidate the PDU port cache
        self._pdu_port = None

    def update_fields(self, fields):
        with self.mars.db as mars_db:
            db_dut = mars_db.duts[self.id]

            updated_fields = set()
            for k, v in fields.items():
                if getattr(self.db_dut, k, None) == v:
                    continue

                if k == 'is_retired':
                    if self.is_retired and not v:
                        self.quick_check_queued.set()

                setattr(db_dut, k, v)

                updated_fields.add(k)

            if len(updated_fields) > 0:
                self.config_changed(db_dut)

    @property
    def _executor_socket_path(self):
        dut_id = self.id.replace(":", "_")
        return f"/var/run/executor/{dut_id}.sock"

    @property
    def _executor_lock_path(self):
        return f"{self._executor_socket_path}.lock"

    def _executor_query(self, path, method="get"):
        session = requests_unixsocket.Session()
        r = getattr(session, method)(f'http+unix://{self._executor_socket_path.replace("/", "%2F")}{path}')
        return r

    @property
    def state(self):
        def job_process_state():
            try:
                r = self._executor_query("/api/v1/state")
                if r.status_code == 200:
                    q = r.json()
                    state = q.get("state")
                    try:
                        return DUTState[state]
                    except KeyError:
                        return DUTState.UNKNOWN
                else:
                    return DUTState.QUEUED
            except (requests.exceptions.ConnectionError, FileNotFoundError):
                # Verify that the lock is not held by anyone before declaring
                # the machine is IDLE
                try:
                    with open(self._executor_lock_path, "r") as f:
                        lock_fd(f, attempts=5)

                        # The lock was available, the machine can be considered IDLE
                        return DUTState.IDLE
                except BlockingIOError:
                    # The lock was not available, the machine may still be in use
                    return DUTState.BORKED
                except FileNotFoundError:
                    # The lock file does not exist, the machine must be idle
                    return DUTState.IDLE

            return DUTState.IDLE

        state = job_process_state()

        if state == DUTState.IDLE:
            # If the state is IDLE, check what state we should give it
            if self.pdu_port is None:
                return DUTState.WAIT_FOR_CONFIG
            elif self.is_retired:
                return DUTState.RETIRED
            elif not self.ready_for_service:
                return DUTState.TRAINING
            elif self.quick_check_queued.is_set():
                return DUTState.QUICK_CHECK
        elif self.sergent_hartman.state == SergentHartmanState.ENROLLING:
            return DUTState.TRAINING
        elif self.sergent_hartman.state == SergentHartmanState.QUICK_CHECK:
            return DUTState.QUICK_CHECK

        return state

    def start_job(self, job_request, quiet=False):
        if self.pdu_port is None:
            raise ValueError("Can't start a job until the PDU port is properly setup")

        # Create the configuration for the run
        config = {
            "executor_job_version": 1,
            "mars_db": asdict(self.mars_db),
            "job_request": asdict(job_request),
            "machine_id": self.id,
            "pdu_port_stats": PduPortStats.from_pdu_port(self.pdu_port).serialize()

        }

        # Compute the list of file descriptors to pass to the next client
        if job_request.job_bucket_initial_state_tarball_file_fd > 0:
            pass_fds = [job_request.job_bucket_initial_state_tarball_file_fd]
        else:
            pass_fds = []

        # Execute the job
        proc = Popen(["executor", "run-job", "-s", self._executor_socket_path,
                      "-l", self._executor_lock_path],
                     stdin=PIPE, stdout=sys.stdout if not quiet else DEVNULL,
                     stderr=sys.stderr if not quiet else DEVNULL,
                     pass_fds=pass_fds)
        yaml.dump(config, proc.stdin, sort_keys=False, encoding='utf-8')
        proc.stdin.flush()
        proc.stdin.close()

        # Wait for 15s for the unix socket to appear, or kill the process
        start = time.monotonic()
        while time.monotonic() - start < 15:
            try:
                self._executor_query("/api/v1/state")
                return proc
            except requests.exceptions.ConnectionError:
                pass

        # Seems like the start-up process did not work, kill the process!
        proc.kill()
        proc.wait()
        raise ValueError("The job process failed to start")

    def cancel_job(self):
        try:
            r = self._executor_query("/api/v1/job/cancel", method="post")
            return r.status_code == 200
        except requests.exceptions.ConnectionError:
            # Nothing to do
            return True

    def boot_config_query(self, platform=None, buildarch=None, bootloader=None):
        try:
            url = f"/api/v1/boot/config?platform={platform}&buildarch={buildarch}&bootloader={bootloader}"
            q = self._executor_query(url)
            if q.status_code == 200:
                return BootConfig(**q.json())
        except Exception:
            traceback.print_exc()
            return None

    def run(self):
        while not self.stop_event.is_set():
            try:
                if self.state == DUTState.TRAINING and self.sergent_hartman.state == SergentHartmanState.IDLE:
                    self.sergent_hartman.set_state(SergentHartmanState.ENROLLING)
                elif self.state == DUTState.QUICK_CHECK and self.sergent_hartman.state == SergentHartmanState.IDLE:
                    self.sergent_hartman.set_state(SergentHartmanState.QUICK_CHECK)
                elif self.sergent_hartman.state == SergentHartmanState.ENROLLING and (self.ready_for_service or
                                                                                      self.is_retired):
                    # Cancel enrollment
                    self.sergent_hartman.set_state(SergentHartmanState.IDLE)

                if self.sergent_hartman.is_active:
                    self.sergent_hartman.execute_next_task(stop_event=self.stop_event)

                    # Check if we are done
                    if self.sergent_hartman.result is not None and not self.stop_event.is_set():
                        if self.sergent_hartman.state == SergentHartmanState.ENROLLING:
                            self.ready_for_service = self.sergent_hartman.result
                        elif self.sergent_hartman.state == SergentHartmanState.QUICK_CHECK:
                            if not self.sergent_hartman.result:
                                self.ready_for_service = False
                            self.quick_check_queued.clear()

                        self.sergent_hartman.set_state(SergentHartmanState.IDLE)

            except Exception:
                traceback.print_exc()

            # Wait for a second before starting a new round of testing
            self.stop_event.wait(1)
